@@ -2,6 +2,7 @@ import os, sys, time, traceback
 import smtplib
 from boto.utils import ShellCommand, get_ts
 import boto
+import boto.utils
 
 class ScriptBase:
 
@@ -13,24 +14,7 @@ class ScriptBase:
             boto.config.read(config_file)
 
     def notify(self, subject, body=''):
-        to_string = boto.config.get_value('Notification', 'smtp_to', None)
-        if to_string:
-            try:
-                from_string = boto.config.get_value('Notification', 'smtp_from', 'boto')
-                msg = "From: %s\n" % from_string
-                msg += "To: %s\n" % to_string
-                msg += "Subject: %s\n\n" % subject
-                msg += body
-                smtp_host = boto.config.get_value('Notification', 'smtp_host', 'localhost')
-                server = smtplib.SMTP(smtp_host)
-                smtp_user = boto.config.get_value('Notification', 'smtp_user', '')
-                smtp_pass = boto.config.get_value('Notification', 'smtp_pass', '')
-                if smtp_user:
-                    server.login(smtp_user, smtp_pass)
-                server.sendmail(from_string, to_string, msg)
-                server.quit()
-            except:
-                boto.log.error('notify failed')
+        boto.utils.notify(subject, body)
 
     def mkdir(self, path):
         if not os.path.isdir(path):
@@ -46,9 +30,11 @@ class ScriptBase:
     def run(self, command, notify=True, exit_on_error=False):
         self.last_command = ShellCommand(command)
         if self.last_command.status != 0:
-            boto.log.error(self.last_command.output)
+            boto.log.error('Error running command: "%s". Output: "%s"' % (command, self.last_command.output))
             if notify:
-                self.notify('Error encountered', self.last_command.output)
+                self.notify('Error encountered', \
+                        'Error running the following command:\n\t%s\n\nCommand output:\n\t%s' % \
+                        (command, self.last_command.output))
             if exit_on_error:
                 sys.exit(-1)
         return self.last_command.status
